#include <sys/types.h>
#include <sys/stat.h>

#include <dirent.h>
#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#ifdef __FreeBSD__
# define NAMLEN(dp) dp->d_namlen
#else
# define NAMLEN(dp) strlen(dp->d_name)
#endif

static struct option longopts[] = {
  { "verbose", no_argument, NULL, 'v' },
  { "version", no_argument, NULL, 'V' },
  { NULL }
};

int verbosity = 0;

void
usage(const char *progname) {
  fprintf(stderr, "usage: %s [-vV] source_dir target_dir\n", progname);
  exit(EXIT_FAILURE);
}

char *
mk_filename(const char *dir,
    const int dir_len,
    const char *name,
    const int name_len,
    int *filename_len)
{
  char *res;

  *filename_len = dir_len + /* strlen("/") */ 1 + name_len;
  if ((res = malloc(*filename_len + 1))) {
    sprintf(res, "%s/%s", dir, name);
  }

  return(res);
}

int
link_dir(const char *source_dir,
    const int source_dir_len,
    const char *target_dir,
    const int target_dir_len)
{
  DIR *dirp;
  struct dirent *dp;
  int err = 0;

  if ((dirp = opendir(source_dir))) {
    while ((dp = readdir(dirp))) {
      if (strcmp(dp->d_name, ".") && strcmp(dp->d_name, "..")) {
        int source_file_len;
        int target_file_len;
        char *source_file;
        char *target_file;

        if ((source_file = mk_filename(source_dir,
                source_dir_len,
                dp->d_name,
                NAMLEN(dp),
                &source_file_len))) {
          if ((target_file = mk_filename(target_dir,
                  target_dir_len,
                  dp->d_name,
                  NAMLEN(dp),
                  &target_file_len))) {

            switch (dp->d_type) {
              case DT_DIR:
                err |= link_dir(source_file,
                    source_file_len,
                    target_file,
                    target_file_len);
                break;
              case DT_REG:
                {
                  struct stat source_stat, target_stat;

                  if ((stat(source_file, &source_stat) == 0) &&
                      (stat(target_file, &target_stat) == 0)) {
                    if ((source_stat.st_ino != target_stat.st_ino) &&
                        (source_stat.st_size == target_stat.st_size) &&
                        (source_stat.st_mtime == target_stat.st_mtime)) {
                      if (verbosity) fprintf(stderr, "%s -> %s\n",
                          source_file,
                          target_file);

                      if (unlink(target_file) == 0) {
                        if (link(source_file, target_file) == -1) {
                          perror("link");
                          err = 1;
                        }
                      } else {
                        perror("unlink");
                        err = 1;
                      }
                    }
                  }
                }
                break;
            } /* switch */

            free(target_file);
          }
          free(source_file);
        } else {
          fprintf(stderr, "Out of memory!\n");
          err = 1;
        }
      } /* if not("." or "..") */
    } /* while readdir() */
    closedir(dirp);
  } else {
    perror("opendir");
    err = 1;
  }

  return(err);
}


int
main(int argc, char *argv[])
{
  int ch;
  struct stat source_stat, target_stat;
  int err = EXIT_SUCCESS;
  const char *progname = argv[0];

  while ((ch = getopt_long(argc, argv, "vV", longopts, NULL)) != -1) {
    switch (ch) {
      case 'v':
        verbosity += 1;
        break;
      case 'V':
        fprintf(stdout, "%s %s (build: %s %s)\n", PROGNAME, VERSION, __DATE__, __TIME__);
        exit(err);
        break;
      default:
        usage(progname);
    }
  }

  argc -= optind;
  argv += optind;

  if (argc != 2) {
    usage(progname);
  }

  if ((stat(argv[0], &source_stat) == 0) &&
      (stat(argv[1], &target_stat) == 0)) {
    if (source_stat.st_dev != target_stat.st_dev) {
      fprintf(stderr, "source and targer directories have to be on the same device.\n");
      err = EXIT_FAILURE;
    } else {
      err = link_dir(argv[0], strlen(argv[0]), argv[1], strlen(argv[1]));
    }
  } else {
    perror("stat");
    err = EXIT_FAILURE;
  }

  exit(err);
}
